#!/usr/bin/perl

# Program:	tokenize_man_src.perl
# Written by:	Dave Graff, LDC
#		(TokenizeGB subroutine written by Xiaoyi Ma, modified by dave graff
# Purpose:	filter tipster-style sgml source data into tokenized "docset" data

# This version is for Mandarin data sources, and handles special properties of
# Mandarin newswire and transcript files

$Usage = 
    "Usage:  $0 fileid\n".
    "  Reads SGML from file ./src_sgm/fileid.src_sgm\n".
    "  Creates tokenized file in ./tkn/fileid.tkn,\n".
    "  Creates boundary table file in ./tkn_bnd/fileid.tkn_bnd\n";

die $Usage 
    if ( $#ARGV != 0 || ! -d "tkn" || ! -d "tkn_bnd" );

$file = shift;
if ( $file =~ /((1998\d{4}_(\d{2})(\d{2}))_\d{4}_([A-Z]+)_[A-Z]+)/i ) {
    $fileid = $1;
    $colldate = $2;
    $bgntime = $3 . ":" . $4;
    $source = $5;
    $file_bgnsec = &time2sec( $bgntime );
} else {
    die "\nUnsuitable file name\n\n$Usage";
}

open( INP, "<src_sgm/$fileid.src_sgm" ) || 
    die "can't open file src_sgm/$fileid.src_sgm for input\n\n$Usage";
open( TKN, ">tkn/$fileid.tkn" ) || 
    die "can't open file tkn/$fileid.tkn for output\n\n$Usage";
open( TBL, ">tkn_bnd/$fileid.tkn_bnd" ) || 
    die "can't open file tkn_bnd/$fileid.tkn_bnd for output\n\n$Usage";

$srctype = ( $source eq "VOA" ) ?  "TRANSCRIPT" : "NEWSWIRE";

print TBL
    "<BOUNDSET type=$srctype fileid=$fileid collect_date=$colldate " .
    "collect_src=$source src_lang=MANDARIN content_lang=NATIVE>\n";
print TKN
    "<DOCSET type=$srctype fileid=$fileid collect_date=$colldate " .
    "collect_src=$source src_lang=MANDARIN content_lang=NATIVE>\n";

$intext = $inannot = $findDL = $recid = 0;

while (<INP>)
{
    if ( ! $intext ) {
	if ( /<DOCNO> (\S+) / ) {
	    $docno = $1;
	    $start = $recid + 1;
	    $use = 1;
	} 
	elsif ( /<DOCTYPE> (\S+) / ) {
	    $doctype = $1;	                    
	    if ( $srctype eq "NEWSWIRE" && $doctype eq "MISCELLANEOUS" ) {
		warn "$fileid:$.:miscellaneous text found in newswire SGML\n";
		$use = 0;
	    }
	}
	elsif ( /<DATE_TIME> \S+ (\S+) / ) {
	    $time = $1;
	    $bgnsec = &time2sec( $time ) - $file_bgnsec;
	}
	elsif ( /<TEXT>/ ) {
	    $intext = $use;	# $intext is 0 for MISC. NEWSWIRE docs, 1 otherwise
	    $findDL++ if ( $srctype eq "NEWSWIRE" ); # watch for "dateline" strings
	}
	elsif ( /<END_TIME> \S+ (\S+) / ) {
	    $time = $1;
	    $endsec = &time2sec( $time ) - $file_bgnsec;
	}
	elsif ( /<\/DOC>/ ) {
	    $entry = sprintf( "<BOUNDARY docno=%s doctype=%s", $docno, $doctype );
	    $entry .= sprintf(  " Bsec=%4.2f Esec=%4.2f", $bgnsec, $endsec )
		if ( $srctype ne "NEWSWIRE" );
	    $entry .= " Brecid=$start Erecid=$recid" if ( $start>0 && $start<=$recid );
	    print TBL "$entry>\n";
	}
	next;
    }
    elsif ( $inannot ) {
	$inannot = 0 if ( /<\/ANNOTATION>/ );
	next;
    }
    else {
	if ( /<ANNOTATION>/ ) {
	    $inannot = 1 unless ( /<\/ANNOTATION>/ );
	    next;
	}
	if ( /<\/TEXT>/ ) {
	    $intext = 0;
	    next;
	}
	next if ( /^<.+>$/ );

	if ( $findDL > 0 ) {   # check for newswire byline:
	    if ( $source eq "XIN" && /յ/ ) {
		$_ = $';
	    }
	    if ( /^[ ]*.*?/ ) {  # initial (1- or 2-byte) spaces, followed by
		$_ = $';	      # 2-byte open-paren, GB text, 2-byte close-paren
	    }
	    $findDL = 0;
	}
	elsif ( $source eq "XIN" && /(꣩)/ ) {
	    s/$1//;
	}
	elsif ( $source eq "ZBN" && /^[ ]{2,}.*?/ ) { # ZBN sometimes has multiple
	    $_ = $';					   # bylines in one story
	}
	@tokens = &tokenizeGB($_);
	foreach $t ( @tokens ) 
	{
	    next if ( $t =~ /^\s*$/ );
	    $recid++;
	    print TKN "<W recid=$recid> $t\n";
	}
    }
}

print TKN "</DOCSET>\n";
print TBL "</BOUNDSET>\n";

close TKN;
close TBL;


sub time2sec
{
    ( $hr, $mi, $sec ) = split( /:/, $_[0] );
    $sec += $hr * 3600 + $mi * 60;
    return $sec;
}


sub tokenizeGB
{
    local($_) = @_;
    my @tkns = ();

# Here are the relevant character patterns to watch for:

    $gb = "[\xa1-\xff][\xa1-\xff]";     # 2-byte patterns with 8th-bit set
    $truegb = "[\xa1-\xfe][\xa1-\xfe]"; # printable 2-byte patterns
    $gbasc = "\xa3[\xa1-\xfe]";         # GB 2-byte versions of alphanumerics, brackets, etc
    $gbspace = "\xa1\xa1";              # GB 2-byte space
    $gbdingbats = "\xa1[\xe8-\xfe]";    # GB "pictorials", e.g. paragraph bullet symbols
    $ascii = "[\x21-\x7e]+";            # single-byte ascii

# First, isolate printable ascii strings as space-separated tokens;
# this will set contiguous sequences of single-byte alphanumerics,
# brackets, punctuation, etc. apart from contiguous sequences of
# two-byte GB characters:

    s/($ascii)/ $1 /g;	# put spaces around all non-space ascii bytes
    split;

# Now go through each string of contiguous characters; ascii strings
# will be output as single token records, while GB strings will be
# split up to treat each 2-byte character as a separate token.

# Withold and report on stderr all bytes that are neither printable
# ascii nor usable as half of a GB character.

    $nbad = 0;
    $bad = "#";
    foreach (@_) {
	if (/^($ascii)$/) {
	    push @tkns, $_;
	} else {
	    $gb2asc = "";
	    until ( length == 0 ) {
		while ( /^($gb)/ ) {
		    $ch = $1;
		    if ( $ch =~ /$gbasc/ ) {
			$och = unpack("H*",substr($ch,1,1));
			$gb2asc .= sprintf( "%c", hex($och) & 0x7f );
		    }
		    elsif ( $ch =~ /$truegb/ ) {
			if ( $gb2asc ne "" ) {
			    push( @tkns, $gb2asc );
			    $gb2asc = "";
			}
			push( @tkns, $ch ) unless ( $ch =~ /($gbspace|$gbdingbats)/ );
		    }
		    s/^$ch//;
		}
		if ( length ) {
		    until ( /^($gb)/ || length == 0 ) {
			$bad .= substr( $_, 0, 1 );
			$nbad++;
			s/^.//;
		    }
		    $bad .= "#";
		}
	    }
	    if ( $gb2asc ne "" ) {
		push( @tkns, $gb2asc );
	    }
	}
    }
    warn "$nbad bad bytes in $fileid $docno near recid $recid: $bad\n" if ( $nbad );
    return @tkns;
}
